////////////////////////////////////////////////////////////////////////////////
// Script:   VMDetection.pas                                                  //
// Language: Delphi Language                                                  //
// Project:  Virtual Machine Detection Unit (VMDU)                            //
// Version:  1.3                                                              //
// Revision: September, 26th 2007                                             //
// Authors:  Zacherl, Daniel Marschall <info@daniel-marschall.de>             //
//                                                                            //
// This unit does check, if your application is running on virtual machine or //
// on a native machine.                                                       //
//                                                                            //
// Please send us suggestions or bug-reports if you have some.                //
//                                                                            //
// This unit supports:                                                        //
// - Windows 95..ME, Windows NT..VISTA                                        //
// - VMWare, VirtualBox, Microsoft Virtual PC, Parallels Workstation, WINE,   //
//   QEMU                                                                     //
//                                                                            //
// We need more tests!                                                        //
// - QEMU                                                                     //
// - WINE                                                                     //
// - VirtualBox                                                               //
// - Other virtual boxes                                                      //
//                                                                            //
// Requirements:                                                              //
// - Delphi 6                                                                 //
// - Windows (Platform unit)                                                  //
//                                                                            //
// Unknown:                                                                   //
// - Does WinNT have LDTR = 0xDEAD0000?                                       //
// - Does a virtual machine give out different IDTLimit's if the parent       //
//   operating system has a different CPU-Architecture (x64, 32-Bit) ?        //
// - /3gb in Win2000?                                                         //
// - Does the assembler part work with IA64/IA32?                             //
// - Detect functions 9x/NT/Linux?                                            //
// - Future: Sandbox-Detection                                                //
////////////////////////////////////////////////////////////////////////////////

{$IFNDEF BCB}

{$DEFINE DEL1UP}
{$IFNDEF VER80}
{$DEFINE DEL2UP}
{$IFNDEF VER90}
{$DEFINE DEL3UP}
{$IFNDEF VER100}
{$DEFINE DEL4UP}
{$IFNDEF VER120}
{$DEFINE DEL5UP}
{$IFNDEF VER130}
{$DEFINE DEL6UP}
{$IFNDEF VER140}
{$DEFINE DEL7UP}
{$ENDIF}
{$ENDIF}
{$ENDIF}
{$ENDIF}
{$ENDIF}
{$ENDIF}

{$ENDIF}

{$IFDEF DEL7UP}
{$WARN UNSAFE_TYPE OFF}
{$WARN UNSAFE_CODE OFF}
{$WARN UNSAFE_CAST OFF}
{$ENDIF}

unit VMDetection platform;

interface

uses
  Windows;

type
  TVMFingerprint = (fpUnknown, fpVMWare, fpVirtualPC, fpVirtualBox, fpNative,
    fpWINE, fpParallelsWorkstation, fpQEMU);

type
  TxDTEntry = record
    GDTBase: DWord;
    IDTBase: DWord;
    LDTRBase: DWord;
    GDTLimit: Word;
    IDTLimit: Word;
  end;

type
  TxDTArray = array of TxDTEntry;
  TFingerprintArray = array of TVMFingerprint;

function IsVirtualMachine(const DetectWINE: Boolean = true): Boolean;
function GetVMFingerprint(xDTEntry: TxDTEntry): TVMFingerPrint; overload;
function GetVMFingerprint(CPU: Cardinal): TVMFingerPrint; overload;
function VMFingerprintToStr(FP: TVMFingerprint): String;
function GetxDTEntry(CPU: Cardinal): TxDTEntry;
function GetxDTArray(var Output: TxDTArray): Cardinal;
function GetFingerprintArray(var Output: TFingerprintArray): Cardinal;

implementation

////////////////////////////////////////////////////////////////////////////////
// PRIVATE FUNCTIONS                                                          //
////////////////////////////////////////////////////////////////////////////////

type
  TxDT = record
    Limit,
    BaseLow,
    BaseHigh: Word;
  end;

function Win9xMe: Boolean;
var
  OSVinfo: TOsVersionInfo;
begin
  ZeroMemory(@OSVinfo,sizeOf(OsVinfo));
  OSVinfo.dwOSVersionInfoSize := sizeof(TOSVERSIONINFO);
  GetVersionEx(OSVinfo);
  Result := OSVInfo.dwPlatformId <> 2;
end;

function GetIDTBase: DWord;
var
  IDT: TxDT;
begin
  asm
    SIDT IDT
  end;
  Result := (IDT.BaseHigh shl 16) or IDT.BaseLow;
end;

function GetIDTLimit: DWord;
var
  IDT: TxDT;
begin
  asm
    SIDT IDT
  end;
  Result := IDT.Limit;
end;

function GetGDTBase: DWord;
var
  GDT: TxDT;
begin
  asm
    SGDT GDT
  end;
  Result := (GDT.BaseHigh shl 16) or GDT.BaseLow;
end;

function GetGDTLimit: DWord;
var
  GDT: TxDT;
begin
  asm
    SGDT GDT
  end;
  Result := GDT.Limit;
end;

function GetLDTRBase: DWord;
var
  Base: Word;
begin
  asm
    sldt Base;
  end;
  Result := $DEAD0000 or Base;
end;

////////////////////////////////////////////////////////////////////////////////
// PUBLIC FUNCTIONS                                                           //
////////////////////////////////////////////////////////////////////////////////

function GetFingerprintArray(var Output: TFingerprintArray): Cardinal;
var
  i: integer;
  Info: TSystemInfo;
begin
  GetSystemInfo(Info);
  SetLength(Output, Info.dwNumberOfProcessors);

  for i := 1 to Info.dwNumberOfProcessors do
  begin
    Output[i-1] := GetVMFingerprint(i);
  end;

  Result := Length(Output);
end;

function GetxDTArray(var Output: TxDTArray): Cardinal;
var
  i: integer;
  Info: TSystemInfo;
begin
  GetSystemInfo(Info);
  SetLength(Output, Info.dwNumberOfProcessors);

  for i := 1 to Info.dwNumberOfProcessors do
  begin
    Output[i-1] := GetxDTEntry(i);
  end;

  Result := Length(Output);
end;

function GetxDTEntry(CPU: Cardinal): TxDTEntry;
var
  Info: TSystemInfo;
begin
  GetSystemInfo(Info);
  if Info.dwNumberOfProcessors >= CPU then
  begin
    SetThreadAffinityMask(GetCurrentThread, CPU);
    Result.GDTBase  := GetGDTBase;
    Result.IDTBase  := GetIDTBase;
    Result.LDTRBase := GetLDTRBase;
    Result.GDTLimit := GetGDTLimit;
    Result.IDTLimit := GetIDTLimit;
  end
  else
  begin
    Result.GDTBase  := $0;
    Result.IDTBase  := $0;
    Result.LDTRBase := $0;
    Result.GDTLimit := 0;
    Result.IDTLimit := 0;
  end;
end;

function GetVMFingerprint(xDTEntry: TxDTEntry): TVMFingerprint;
begin
  Result := fpUnknown;

  { if (xDTEntry.IDTLimit       = 2048) then
  begin
    // Buggy version of Norman Sandbox
    // http://www.heise.de/newsticker/result.xhtml?url=/newsticker/meldung/86041&words=7FF%20Hardware%20Software&T=vmware%207ff
    Result := fpNormanSandbox;
    exit;
  end; }

  if (xDTEntry.LDTRBase  = $DEAD4060) and
     (xDTEntry.GDTLimit  =     16687) and
     (xDTEntry.IDTBase   = $FFC18000) and
     (xDTEntry.GDTBase   = $FFC07000) then
  begin
    Result := fpVMWare;
    exit;
  end;

  if (xDTEntry.GDTLimit  =       255) and
  // (xDTEntry.IDTLimit  =      2047) and
     (xDTEntry.LDTRBase  = $DEAD0088) then
  begin
    Result := fpWINE;
    exit;
  end;

  if (xDTEntry.GDTLimit  =     65535) and
  // (xDTEntry.IDTLimit  =      2047) and
     (xDTEntry.LDTRBase  = $DEADFFA8) then
  begin
    Result := fpVirtualPC;
    exit;
  end;

  if (xDTEntry.GDTLimit  =     65535) and
  // (xDTEntry.IDTLimit  =      2047) and
     (xDTEntry.LDTRBase  = $DEADFF5B) then
  begin
    Result := fpParallelsWorkstation;
    exit;
  end;

  if (xDTEntry.GDTLimit  =       503) and
  // (xDTEntry.IDTLimit <=      2047) and
     (xDTEntry.LDTRBase  = $DEAD00D0) then
  begin
    Result := fpQEMU;
    exit;
  end;

  if                    (xDTEntry.GDTLimit  =     65535)  and
  //                    (xDTEntry.IDTLimit  =      2047)  and
     ((    Win9xMe) and (xDTEntry.LDTRBase  = $DEAD00B8)  or
      (not Win9xMe) and (xDTEntry.LDTRBase  = $DEAD0000)) then
  begin
    Result := fpVirtualBox;
    exit;
  end;

  if ((    Win9xMe) and (xDTEntry.LDTRBase  = $DEAD00B8)
                    and (xDTEntry.GDTLimit <=      4095)
                    and (xDTEntry.IDTLimit  =       767)) or
     ((not Win9xMe) and (xDTEntry.LDTRBase  = $DEAD0000)
                    and (xDTEntry.GDTLimit <=      1023)
                    and (xDTEntry.IDTLimit <=      4095)) then
  begin
    Result := fpNative;
    exit;
  end;
end;

function GetVMFingerprint(CPU: Cardinal): TVMFingerprint;
var
  xDTEntry: TxDTEntry;
begin
  xDTEntry := GetxDTEntry(CPU);
  Result := GetVMFingerprint(xDTEntry);
end;

function VMFingerprintToStr(FP: TVMFingerprint): String;
begin
  case FP of
    fpUnknown:              Result := 'Unknown';
    fpVMWare:               Result := 'VMWare';
    fpVirtualPC:            Result := 'VirtualPC';
    fpVirtualBox:           Result := 'VirtualBox';
    fpWINE:                 Result := 'WINE';
    fpNative:               Result := 'Native';
    fpParallelsWorkstation: Result := 'Parallels Workstation';
    fpQEMU:                 Result := 'QEMU';
  end;
end;

function IsVirtualMachine(const DetectWINE: Boolean = true): Boolean;
var
  i: integer;
  Info: TSystemInfo;
begin
  Result := true;
  GetSystemInfo(Info);

  for i := 1 to Info.dwNumberOfProcessors do
  begin
    if ((    DetectWINE) and  (GetVMFingerprint(i) = fpNative)) or
       ((not DetectWINE) and ((GetVMFingerprint(i) = fpNative)
                          or  (GetVMFingerprint(i) = fpWINE)))  then
    begin
      Result := false;
      Break;
    end;
  end;
end;

end.
